#include <mcs51/8051.h>


#define OSC 100000000.0  //DDS clock frequency


void lcd_command(unsigned char cmd);
void lcd_putchar(char c);
void to_string(unsigned long n, char r[]);
void AD_Write(unsigned long spiData);
void Delay();
void Delay1();
void Delay2();


__sbit __at (0xA0)  LCDE;       //LCD En signal
__sbit __at (0xA1)  FSYNC;      //DDS FSYNC signal
__sbit __at (0xA2)  PSELECT;    //DDS PSELECT signal
__sbit __at (0xA3)  FSELECT;    //DDS FSELECT signal
__sbit __at (0xA4)  SCKF;       //DDS Shift clock
__sbit __at (0xA5)  SDATA;      //DDS data signal
__sbit __at (0xA6)  RESET;      //DDS reset signal
__sbit __at (0xA7)  INH;        //DDS DAC inhibit signal

__sbit __at (0xB0)  RX;     //Serial RX pin
__sbit __at (0xB1)  TX;     //Serial TX pin
__sbit __at (0xB2)  RA;     //Rotary encoder A pin
__sbit __at (0xB3)  RBtn;   //Rotary encoder button
__sbit __at (0xB4)  RB;     //Rotary encoder B pin
__sbit __at (0xB5)  MBtn;   //push button switch
__sbit __at (0xB6)  EP;     //External pulse
__sbit __at (0xB7)  ET;     //Sweep trigger

unsigned char __at (0x90)  LCD;

enum ScrenModes{Main=0,SweepStart=1,SweepStop=2,SweepSpeed=3,Sweeping=4};  //Menu modes

unsigned long input=1000;
float frequency=1000;
unsigned long scale=100;
unsigned char waveform=0;
unsigned char Screen=0;
volatile char updateDisplay=0;

const float mult[8]={//steps per decade
    1.0002305238,   //1000
    1.00461579,     //500
    1.011579454,    //200
    1.023292992,    //100
    1.047128548,    //50
    1.122018454,    //20
    1.258925412,    //10
    1.584893192     //5
    };

long start=1000; //Sweep start frequency
long stop=2000; //Sweep stop frequency
unsigned char speed=1; //Sweep speed


#define lcd_set_xy(x, y) (lcd_set_xy_raw((x) + ((y) << 8)))

#define LCD_COMMAND_WR 0xFE00
#define LCD_STATUS_RD 0xFE01
#define LCD_DATA_WR 0xFE02
#define LCD_DATA_RD 0xFE03

#define LCD_CLEAR_CMD 0x01
#define LCD_HOME_CMD 0x02
#define LCD_ON_CMD 0x0C
#define LCD_SHIFT_CMD 0x10
#define LCD_CONFIG_CMD 0x38
#define LCD_SECOND_LINE 40 | 0x80


#define AD9834_REG_CMD		(0 << 14)
#define AD9834_REG_FREQ0	(1 << 14)
#define AD9834_REG_FREQ1	(2 << 14)
#define AD9834_REG_PHASE0	(6 << 13)
#define AD9834_REG_PHASE1	(7 << 13)

#define AD9834_B28			(1 << 13)
#define AD9834_HLB			(1 << 12)
#define AD9834_FSEL0		(0 << 11)
#define AD9834_FSEL1		(1 << 11)
#define AD9834_PSEL0		(0 << 10)
#define AD9834_PSEL1		(1 << 10)
#define AD9834_CMD_PIN		(1 << 9)
#define AD9834_CMD_SW		(0 << 9)
#define AD9834_RESET		(1 << 8)
#define AD9834_SLEEP1		(1 << 7)
#define AD9834_SLEEP12		(1 << 6)
#define AD9834_OPBITEN		(1 << 5)
#define AD9834_SIGN_PIB		(1 << 4)
#define AD9834_DIV2			(1 << 3)
#define AD9834_MODE			(1 << 1)


#define IEREG 0b10010001

const float F_CONST=268435456.0/OSC;

void AD9834_Setinput(float freq)
{
	unsigned short freqHi = 0x4000;
	unsigned short freqLo = 0x4000;
	unsigned long val;

	val=(unsigned long)(freq*F_CONST);

	freqHi |= (val & 0xFFFC000) >> 14 ;
	freqLo |= (val & 0x3FFF);
	if(waveform==0) //Sine
	{
	    INH=1;
        AD_Write(AD9834_B28 | AD9834_OPBITEN | AD9834_DIV2);
	}
    else if(waveform==1) //Triangle
    {
        INH=1;
        AD_Write(AD9834_B28 | AD9834_MODE | AD9834_DIV2);
    }
    else if(waveform>1) //Square wave
    {
        AD_Write(AD9834_B28 | AD9834_SLEEP12 | AD9834_OPBITEN | AD9834_DIV2);
        INH=0;
    }

	AD_Write(freqLo);
	AD_Write(freqHi);
}

void print(char str[]) //Print to LCD
{
    unsigned char i;
    for(i=0;str[i];i++)
        lcd_putchar(str[i]);

    return;
}

void Display() //Update display
{
    char txt[16],i;
    unsigned long sc;

    IE=0;
    lcd_command(12);

    if(Screen==Main) //Prepare the frequency value for printout
        to_string(input,txt);

    if(Screen==SweepStart) //Prepare the sweep start value for printout
        to_string(start,txt);

    if(Screen==SweepStop) //Prepare the sweep stop value for printout
        to_string(stop,txt);

    if(Screen==SweepSpeed) //Prepare the sweep speed value for printout
        to_string(speed,txt);

    if(updateDisplay>1) //If entire display needs to be updated
    {

        lcd_command(LCD_CLEAR_CMD);
        lcd_command(LCD_HOME_CMD);


        if(Screen==Main)
            print("Wave:");

        if(Screen==SweepStart)
            print("Start:");

        if(Screen==SweepStop)
            print("Stop:");

        if(Screen==SweepSpeed)
           print("Speed:");

        if(Screen==Sweeping)
        {
            print("Sweeping...");
            return;
        }

        print(" ");
        if(waveform==0)
            print("SIN");
        if(waveform==1)
            print("TRI");
        if(waveform==2)
            print("SQW");

        print(" Freq:");
    }

    updateDisplay=0;//Clear update flag
    lcd_command(LCD_SECOND_LINE);
    for(i=0;txt[i];i++);//Get txt string length
    for(;i<7;i++)
        lcd_putchar('_');//Print underscores for leading zeroes

    print(txt);//print the frequency value
    if(Screen<3)
    {
        print(" Hz ");//Print the Hz unit
        lcd_command(16);//Go back 4 times
        lcd_command(16);
        lcd_command(16);
        lcd_command(16);
    }
    lcd_command(16);//Go back one position
    for(sc=scale;sc>1;sc/=10)
        lcd_command(16);//Go back to the selected position
    lcd_command(14);//Enable cursor
    IE=IEREG;
}

void left()//Rotary encoder rotated counter clockwise, reduce value
{
    while(scale>=10)
    {
        if(scale>=input)
            scale/=10;
        else
            break;
    }

    input-=scale;
    if(input<1)
        input=1;
    if(Screen<SweepSpeed)
    {
        frequency=input;
        AD9834_Setinput(frequency);
    }
}

void right()//Rotary encoder rotated clockwise, increase value
{
    input+=scale;
    if(input>20000000) //Max frequency limit
        input=20000000;
    if(Screen<SweepSpeed)
    {
        frequency=input;
        AD9834_Setinput(frequency);
    }
}
unsigned char ec=0; //Debouncer counter
__sbit rotation=0; //Rotation direction bit

void Encoder() __interrupt(0)//Rotary encoder interrupt
{
    TCON=0;
    IE=0;
    if(Screen<Sweeping)
    {
        for(ec=3;ec;ec--)//Check the value multiple times to debounce the signal
            if(RB)
                rotation=1;

        for(ec=0;ec<5;ec++)//Loop until the signal stabilizes
            if(!RA||!RB)
                ec=0;


        if(rotation)
            left();
        else
            right();
        updateDisplay|=1;

        if(Screen==SweepStart)
            start=input;

        if(Screen==SweepStop)
        {
            stop=input;
            if(stop<(start+1)) //Limit to start value
                stop=input=start+1;
        }

        if(Screen==SweepSpeed)
        {
            scale=1;
            if(input>8) //Limit to 8 max
                input=8;
            speed=input;
        }
        rotation=0;
    }
    IE=IEREG;
    TCON=5;
}

volatile unsigned long parse=0;
char in;

void Serial(void) __interrupt(4)//Serial port parser
{
    IE=0;
    in=SBUF;

    if(in<='9')//If number
        if(in>='0')//Split if-s to save space
        {
            parse*=10;
            parse+=in&7;//add digit
        }



    if(in<16)//If new line character, apply the new frequency and waveform
    {
        AD9834_Setinput(parse);
        in='c';//Set to clear
    }

    if(in=='c')//If clear command, set input to 0
        parse=0;
    else//Else waveform select
    {
        if(in>='A')//If letter
            waveform=(in&3)-1;//decode value, A=Sine, B=Triangle, C&D=Square wave, and repeats over for all other letters
    }

    RI=0;
    IE=IEREG;
}

unsigned char rbtco=0,rbtstat=0;

void main(void)
{
    unsigned short co=0; //Initialisation section
    RESET=1;
    INH=1;
    FSYNC=1;
    SCKF=1;
    SDATA=0;
	LCDE=1;
    PSELECT=0;
    FSELECT=0;
	MBtn=1;
	ET = 0;
	Delay();
	RESET=0;
	lcd_command(LCD_CONFIG_CMD); //Setup LCD
	lcd_command(LCD_ON_CMD);
	lcd_command(LCD_SHIFT_CMD);

    AD_Write(AD9834_REG_CMD | AD9834_RESET | AD9834_OPBITEN | AD9834_DIV2); //Setup the DDS IC

    lcd_command(LCD_HOME_CMD);

    updateDisplay=3;
    AD9834_Setinput(input);


    TMOD = 0x20;//Set up serial port
	TH1  = 0xFD;
	SCON = 0x50;
	TR1  = 1;


    IE=IEREG;

	while(1)
	{

	    if(RBtn&&Screen<SweepSpeed)//If rotary encoder button is pressed, change the input scale
        {
            if(rbtco<30)//Button debouncing
                rbtco++;
            else
            {
                if(!rbtstat)
                {
                    rbtstat=1;
                    scale*=10;
                    if(scale>1000000)
                        scale=1;
                    updateDisplay|=1;
                }
            }
        }
        else
        {
            if(rbtco>1)//Button debouncing
                rbtco--;
            else
                rbtstat=0;
        }

	    if(updateDisplay)//Check if display needs updating
        {
            Display();
            Delay2();
            Delay2();
            Delay2();
            Delay2();
        }


	    if(!MBtn)//Tact switch button
        {
            for(co=0;co<40&&!MBtn;co++)
                Delay2();
            if(co<40&&Screen<3)//If pressed quickly change the waveform
            {
                waveform=(waveform+1)%3;
                AD9834_Setinput(input);
                updateDisplay=3;
            }
            else//If held down, change the mode
            {
                Screen++;
                switch(Screen)
                {
                    case SweepStart://Sweep start frequency
                        start=input=20;
                        break;
                    case SweepStop://Sweep stop frequency
                        stop=input=20000;
                        break;
                    case SweepSpeed://Sweep speed 1-8
                        speed=input=4;
                        break;
                    case Sweeping://Sweeping mode
                        frequency=start;
                        updateDisplay=3;
                        Display();
                        IE=0;
                        speed--;
                        stop++;
                        while(RBtn)//Loop until button is pressed
                        {
                            frequency*=mult[speed]; //Increase the frequency
                            if(frequency>stop)
                            {
                                ET=1; //Trigger signal
                                frequency=start; //Restart the frequency
                                AD9834_Setinput(frequency);
                            }
                            else
                                AD9834_Setinput(frequency);
                            ET=0;
                        }
                        while(!RBtn);//Wait for button release
                        Screen=0;
                        speed++;
                        stop--;
                        IE=IEREG;
                        break;
                }
                updateDisplay=3;
            }

            Display();
            while(!MBtn);
            Delay2();
	    }

	}

}

void Delay(void)
{
    char i;
    for(i=0;i<10;i++);
}

void Delay1(void)
{
    char i;
    for(i=0;i<100;i++);
}

void Delay2(void)
{
    unsigned short i;
    for(i=0;i<2000;i++);
}

void AD_Write(unsigned long spiData) //Bitbang write to DDS IC
{
    unsigned char i=16;
    FSYNC=0;
    do
    {
        i--;
        SCKF=1;
        SDATA=(spiData>>i)&1;
        SCKF=0;
    }while(i);
    SCKF=1;
    FSYNC=1;
}


void lcd_command(unsigned char cmd) //Send LCD command
{
	P1 = cmd;
	SDATA=0;
	LCDE=0;
	LCDE=1;
	Delay1();

}

void lcd_putchar(char c) //Write char to LCD
{
	P1 = c;
	SDATA=1;
	LCDE=0;
	LCDE=1;
	Delay1();
}

void strrev(unsigned char *str) //Reverse string
{
	unsigned char i;
	unsigned char j;
	unsigned char a;
	unsigned char len;;
	for(len=0;str[len];len++);

	for(i=0,j=len-1;i<j;i++,j--)
	{
		a = str[i];
		str[i] = str[j];
		str[j] = a;
	}
}

void to_string(unsigned long n, char r[]) //Integer to string
{
    unsigned char i;

    for(i=0;n>9;i++)
    {
        r[i]=(n%10)+'0';
        n/=10;
    }
    r[i++]=n+'0';
    r[i]=0;
    strrev(r);
}



